"""High-level representation of video quality"""
__all__ = ('Quality',)

import re

_match_quality_int = re.compile(r'(\d+)([piK])\s*($|\w+)').match


class Quality(str):
    """str type add-on to represent video quality"""
    __slots__ = ('_i', 'addon', 'units')
    addon: str

    def __init__(self, *_, **__):
        """Sets addon attribute (can contain 'ultra')"""
        _i = _match_quality_int(self)
        if not _i:
            raise ValueError(f'{self!r} is unknown quality.')
        _i, units, self.addon = _i.groups()
        if units == 'K':
            _i += '000'
            units = 'p'
        self.addon = self.addon.casefold()
        self._i = int(_i)
        self.units = units

    def __int__(self):
        """
        returns pixels height
        """
        return self._i

    def __lt__(self, other):
        """Is other quality better then self"""
        if not isinstance(other, self.__class__):
            return super().__le__(other)
        if not (self.addon or other.addon):
            return self._i < int(other)
        return False



"""Inline item datatypes"""
__all__ = ('InlineItem', 'InlineInfo')

from types import EllipsisType
from typing import NamedTuple


class InlineInfo(NamedTuple):
    """Info about inline item (bottom)"""
    year: int
    year_final: int | EllipsisType | None
    'If the film is equal to None, if ongoing, equal `...`'
    country: str
    genre: str


class InlineItem(NamedTuple):
    """Content Inline Item view"""
    url: str
    name: str
    info: InlineInfo
    poster: str
    'Image url'

    @property
    async def player(self):
        """Return a Player Instance"""
        from .. import Player
        return await Player(self.url)



"""High-level representation of video"""
__all__ = ('VideoURL', 'VideoURLs')

from typing import Iterable, SupportsInt

from .quality import Quality
from ...._deobfuscation import clear_trash
from ....api.http import get_response


class _AwaitableStr(str):
    __slots__ = ()

    def __await__(self):
        """Follow redirects in url and return correct video url"""
        return self.__class__((yield from get_response('GET', self, follow_redirects=False).__await__()
                               ).headers.get('Location', str(self)))


class VideoURL(_AwaitableStr):
    """str type add-on to represent video"""
    __slots__ = ()

    @property
    def mp4(self) -> str:
        """URL without ':hls:manifest.m3u8'"""
        return _AwaitableStr(self.removesuffix(':hls:manifest.m3u8'))


class VideoURLs:
    """Class representing video urls"""
    __slots__ = ('raw_data', 'qualities', 'min')

    def __init__(self, data: str | dict):
        """
        :param data:
            type str if raw url from AJAX request
            else type dict[Quality, VideoURLs]
            if not str and not dict raises TypeError
        """
        if isinstance(data, str):
            self.raw_data: dict[Quality, tuple[VideoURL, ...]] = {
                Quality(q): (*(VideoURL(i) for i in u.split(' or ') if i.endswith('.m3u8')),)
                for q, u in (i.removeprefix('[').split(']', 1) for i in clear_trash(data).split(','))
            }
        elif isinstance(data, dict):
            self.raw_data = data
        else:
            raise TypeError(f'got {data!r} (type {type(data)}) but expected type str | dict')
        self.qualities: tuple[Quality, ...] = *sorted(self.raw_data),
        self.min = int(self.qualities[0]) if self.qualities else 1

    @property
    def last_url(self) -> tuple[VideoURL, ...]:
        """Best quality url"""
        return self[-1].raw_data.popitem()[1]

    def __getitem__(self, item: str | SupportsInt | Iterable | slice):
        """
        >>> self[1080]['ultra']
        {'1080p Ultra': '...'}
        >>> self[360:1080, 2160]
        {'360p': '...', '480p': '...', '720p': '...', '2160p': '...'}
        >>> self[720, 1080]
        {'720p': '...', '1080p': '...', '1080p Ultra': '...'}
        """
        if isinstance(item, str):
            item = item.casefold()
            result = {q: v for q, v in self.raw_data.items() if q.addon == item}
        elif isinstance(item, slice):
            supported = {*range(*item.indices(int(self.qualities[-1]) + 1))}
            result = {q: v for q, v in self.raw_data.items() if int(q) in supported}
        elif isinstance(item, Iterable) and item:
            result = {}
            for part in item:
                if part is not None:
                    result |= self[part].raw_data
        elif isinstance(item, int):
            if item < self.min:
                item = self.qualities[item]
                result = {item: self.raw_data[item]}
            else:
                result = {q: v for q, v in self.raw_data.items() if int(q) == item}
        else:
            raise TypeError(f'Invalid type {type(item)}')
        return self.__class__(result)

    def __repr__(self):
        return f'{self.__class__.__qualname__}({self.raw_data!r})'



"""Implementation of search on HDRezka pages (see ..post.page.Page class)"""
__all__ = ('Search',)

from typing import override

from ..post.page import *
from ..url import Request


class Search(Page):
    """AJAX class for HDRezka search"""
    __slots__ = ('_query',)

    def __init__(self, query: str = ''):
        """Preparing query for search"""
        self.query = query.strip()
        super().__init__(self.search_url(self.query))

    @staticmethod
    def search_url(query: str):
        """Returns current query search url"""
        return Request.host_join(f'search/?do=search&subaction=search&q={query}')

    @property
    def query(self) -> str:
        """Current search query"""
        return self._query

    @query.setter
    def query(self, value):
        """Cast value to str and sets"""
        self._query = value if isinstance(value, str) else str(value)
        self.page = self.search_url(self.query)

    @staticmethod
    @override
    def _concat_paginator(url: str) -> str:
        return f'{url}&page={{0}}'

    def __repr__(self):
        return f"{self.__class__.__qualname__}({repr(self.query) if self.query else ''})"



"""Some internal utils"""
__all__ = ('get_any_int', 'parse_help_href', 'hyperlink', 'text', 'attr', 'hyperlinks', 'empty_tag', 'poster_and_soup',
           'sidecover')

import re
from base64 import b64decode
from typing import Any
from urllib.parse import unquote

from bs4 import Tag, BeautifulSoup

from ..info.fields import *
from ..._bs4 import BUILDER
from ...api.http import get_response

empty_tag = BeautifulSoup('<div></div>', builder=BUILDER).find('div')
_find_any_digit = re.compile(r'\d+').findall


def text(tag: Tag | None, default: Any = ''):
    """Get tag text"""
    if tag is None:
        return default
    return tag.text.strip()


def attr(tag: Tag | None, name: str, *, default=None):
    """Get tag attribute"""
    if tag is None:
        return default
    return tag.attrs.get(name, default)


def hyperlinks(tag: Tag) -> tuple[Hyperlink, ...]:
    """Get tag hyperlinks"""
    return *map(hyperlink, tag.find_all('a')),


def get_any_int(value: Any, /, default: int = 0) -> int:
    """Convert Any to int"""
    if isinstance(value, int):
        return value
    if not isinstance(value, str):
        value = str(value)
    if digits := _find_any_digit(value):
        return int(''.join(digits))
    return default


def parse_help_href(href: str) -> str:
    """Returns encoded in base64 url"""
    return unquote(b64decode(href.removeprefix('/help/').removesuffix('/')).decode())


def hyperlink(tag: Tag | None) -> Hyperlink:
    """Returns Hyperlink from tag"""
    if not tag:
        return Hyperlink()
    return Hyperlink(tag.text, tag.attrs.get('href', ''))


def sidecover(tag: BeautifulSoup) -> Poster:
    """Returns sidecover Poster"""
    return (Poster(attr(poster, 'href', default=''), attr(poster.find('img'), 'src', default=''))
            if (poster := tag.select_one('.b-sidecover>a')) else Poster())


async def poster_and_soup(url: str) -> tuple[BeautifulSoup, Poster] | None:
    """Returns soup and poster from url"""
    if not url:
        return None
    soup = BeautifulSoup((await get_response('GET', url)).content, builder=BUILDER)
    return soup, sidecover(soup)



"""Sending AJAX requests"""
__all__ = ('AJAX',)

from typing import SupportsInt

import httpx

from .http import get_response
from ..api.types import *
from ..errors import AJAXFail
from ..url import Request


class AJAX:
    """Class with methods for sending AJAX requests to the API"""
    __slots__ = ()

    @classmethod
    def _check(cls, resp: httpx.Response):
        answer = resp.json()
        if not answer.get('success', True):
            raise AJAXFail(answer.get('message', answer))
        return answer

    @classmethod
    async def _send_data(cls, action: str, **kwargs):
        return cls._check(await get_response('POST', Request.host_join(f'ajax/{action}/'), **kwargs))

    @classmethod
    async def get_cdn_series(cls, data: dict[str, SupportsInt | str]):
        """Request series for one translation"""
        return await cls._send_data('get_cdn_series', data=data)

    @classmethod
    async def get_episodes(cls, id_: SupportsInt | str, translator_id: SupportsInt | str):
        """Request episodes"""
        return await cls.get_cdn_series(
            {'action': 'get_episodes',
             'id': id_,
             'translator_id': translator_id}
        )

    @classmethod
    async def get_stream(cls, id_: SupportsInt | str, translator_id: SupportsInt | str,
                         season: SupportsInt | str, episode: SupportsInt | str) -> APIResponse:
        """Request stream urls"""
        return await cls.get_cdn_series(
            {'action': 'get_stream',
             'id': id_,
             'translator_id': translator_id,
             'season': season,
             'episode': episode}
        )

    @classmethod
    async def get_movie(cls, id_: SupportsInt | str, translator_id: SupportsInt | str):
        """Request movie urls"""
        return await cls.get_cdn_series(
            {'action': 'get_movie',
             'id': id_, 'translator_id': translator_id}
        )

    @classmethod
    async def get_trailer_video(cls, id_: SupportsInt | str) -> TrailerResponse:
        """Request trailer video iframe"""
        return cls._check(await get_response('POST', Request.host_join('engine/ajax/gettrailervideo.php'),
                                             data={'id': id_}))
